import * as anchor from "@project-serum/anchor";
import { ValidateKycData, Gender } from "../types";
import * as countryLookup from "country-code-lookup";
import { ProviderClient } from "../clients/provider-client";
const { createHash } = require("crypto");

export function keypairFromJson(secretKey: number[]): anchor.web3.Keypair {
  return anchor.web3.Keypair.fromSecretKey(Uint8Array.from(secretKey));
}

export function normalizeName(name: string) {
  const removeDiacriticsAndConvertSpecialCharacters = (str: string) => {
    return str
      .normalize("NFD") // Remove diacritics
      .replace(/[\u0300-\u036f]/g, "") // Convert special characters
      .replace(/Đ/g, "D"); // Replace Đ with D
  };

  const capitalizeFirstLetter = (str: string) => {
    return str.charAt(0).toUpperCase() + str.slice(1);
  };

  const trimAndSingleSpace = (str: string) => {
    return str.trim().replace(/\s+/g, " ");
  };

  let normalizedString = removeDiacriticsAndConvertSpecialCharacters(name);
  normalizedString = normalizedString.toLowerCase().trim();
  normalizedString = normalizedString
    .split(" ")
    .map((word) => capitalizeFirstLetter(word))
    .join(" ");
  normalizedString = trimAndSingleSpace(normalizedString);

  return normalizedString;
}

function hashNTimes(string: string | null, salt: string, n: number): string | null {
  if (!string || string === "") return null;
  let hash = createHash("sha256");
  let result = hash.update(string).update(salt).digest("hex");
  for (let i = 1; i < n; i++) {
    let hashInterval = createHash("sha256");
    hashInterval.update(result);
    result = hashInterval.digest("hex");
  }

  return result;
}

export const stringToBytes32 = (str: string | null): number[] => {
  if (!str) return [];
  const encoder = new TextEncoder();
  const encodedString = encoder.encode(str);
  const bytes32 = new Uint8Array(32);
  bytes32.set(encodedString.subarray(0, 32));
  const numbersArray: number[] = Array.from(bytes32, (byte) => byte);
  return numbersArray;
};

export const validateAndHashKycParams = (
  client: ProviderClient,
  kycLevel: number,
  salt: string,
  name: string | null,
  documentId: string | null,
  country: string,
  phone: string,
  dob: Date | null,
  doe: Date | null,
  gender: Gender | null,
): ValidateKycData => {
  // Validate salt
  if (!salt || salt === "") {
    throw Error("Salt is required!");
  }
  if (salt.length > 20) {
    throw Error("Invalid salt length! Cannot exceed 20 characters");
  }
  // Validate phone number
  if (!phone || phone === "") {
    throw Error("Phone number is required!");
  }
  if (phone.length > 255) {
    throw Error("Invalid phone length! Cannot exceed 255 characters");
  }
  // Require parameters for kyc level higher than 1
  if (kycLevel !== 1) {
    if (!name || name === "") {
      throw Error("Name is required!");
    };
    if (!documentId || documentId === "") {
      throw Error("Document ID is required!");
    };
    if (!country || country === "") {
      throw Error("Country is required!");
    };
  }
  let encodedName = null;
  if (name) {
    if (name.length > 255) {
      throw Error("Invalid name length! Cannot exceed 255 characters");
    }
    encodedName = normalizeName(name);
  }
  if (documentId) {
    if (documentId.length > 255) {
      throw Error("Invalid document ID length! Cannot exceed 255 characters");
    }
  }
  if (phone) {
    if (phone.length > 255) {
      throw Error("Invalid phone number length! Cannot exceed 255 characters");
    }
  }
  if (country.length !== 2) {
    throw Error("Invalid country! Please provider valid country code");
  }

  let encodedCountry = country.toLowerCase();
  if (encodedCountry !== "eu" && encodedCountry !== "go" && !countryLookup.byIso(encodedCountry)) {
    throw Error("Invalid country");
  }

  let encodedGender: number[] = [];
  if (gender) {
    if (!Object.values(Gender).includes(gender)) {
      throw Error("Invalid gender");
    }
    encodedGender = stringToBytes32(hashNTimes(gender, salt, client.genderHashRound));
  }

  return {
    encodedName: encodedName
      ? stringToBytes32(hashNTimes(encodedName, salt, client.nameHashRound))
      : [],
    encodedDocumentId: documentId
      ? stringToBytes32(hashNTimes(documentId, salt, client.documentIdHashRound))
      : [],
    encodedCountry,
    encodedPhone: stringToBytes32(hashNTimes(phone, salt, client.phoneHashRound)),
    encodedDob: dob
      ? stringToBytes32(hashNTimes(dob.getTime().toString(), salt, client.dobHashRound))
      : [],
    encodedDoe: doe
      ? stringToBytes32(hashNTimes(doe.getTime().toString(), salt, client.doeHashRound))
      : [],
    encodedGender,
  };
};
